/*-
 * Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: deleted.c,v 1.8 2014/10/03 10:54:03 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "deleted.h"

static int
prio_set(deleted_pid *conf) {
    bool correct;
    int prio;

#   define blin_internal_flags (conf->flags & BLIN_MASK)
    correct = true;
    errno = 0;
    prio = getpriority(PRIO_PROCESS, 0);
    if  ((-1 == prio) && errno) {
        ifBLIN_QW1("getpriority");
        prio = 999;
        correct = false;
    }
    if  (!correct || (prio < conf->prio)) {
        if  (!setpriority(PRIO_PROCESS, 0, conf->prio)) {
            ifBLIN_QW1("setpriority");
        }
        errno = 0;
        prio = getpriority(PRIO_PROCESS, 0);
        if  ((-1 == prio) && errno) {
            ifBLIN_QW1("getpriority");
            prio = 999;
    }   }
    return(prio);
#   undef blin_internal_flags
}

static void
main_loop(deleted_pid *parm, int prio) {
    time_t now, unsleep;
    d_entry *tree;
    struct tm t;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    for (unsleep = 0, tree = NULL; prio <= parm->conf->prio;) {
        parm->conf->res_low = false;                                        /*XXXX need res_low test */
        /* res_low - indicaror RESURSES LOW
         * mean that some request for resource (usually memory)
         * is not satisfied.
         */
        if  (!unsleep) tree = build_tree(parm, "/");
        now = NOW;
        if  (!!tree && !! tree->time) {
            parm->conf->wake_up = tree->time + BONUS_LIFE;
            ifBLIN_QX4("%d", tree->time);
            free_tree(tree);
        } else {
            if  (!unsleep || (parm->conf->fmask.ttl && (unsleep > parm->conf->fmask.ttl))) {
                unsleep = parm->conf->fmask.ttl;
            }
            if  (!unsleep || (parm->conf->dmask.ttl && (unsleep > parm->conf->dmask.ttl))) {
                unsleep = parm->conf->dmask.ttl;
            }
            if  (!unsleep || (parm->conf->lmask.ttl && (unsleep > parm->conf->lmask.ttl))) {
                unsleep = parm->conf->lmask.ttl;
            }
            parm->conf->wake_up = unsleep + now + BONUS_LIFE;
        }
        ifBLIN_QO1 localtime_r(&now, &t);
        ifBLIN_QX1( "           now %04d-%02d-%02d %02d:%02d:%02d %s"
                  , 1900 + t.tm_year, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec, t.tm_zone
                  );
        if  (parm->conf->wake_up > now) {
            ifBLIN_QO1 localtime_r(&parm->conf->wake_up, &t);
            ifBLIN_QX1( ",   wake_up at %04d-%02d-%02d %02d:%02d:%02d %s"
                      , 1900 + t.tm_year, t.tm_mon, t.tm_mday
                      , t.tm_hour, t.tm_min, t.tm_sec, t.tm_zone
                      );
        }
        unsleep = 0;
        if  (parm->conf->wake_up > now) unsleep = sleep(parm->conf->wake_up - now);
        prio = prio_set(parm->conf);
    }
#   undef blin_internal_flags
}

static time_t
get_time(deleted_pid *parm, babolo_opts *bos) {
    time_t t, tt = 0;
    const char *in;

#   define blin_internal_flags (parm->flags & BLIN_MASK)
    if  (!(in = babolo_getoptsarg(bos))) {
        ifBLIN_QW0("getoptsarg -%c", bos->c);
        goto out;
    }
    for (tt = 0, t = 0; *in; ++in) {
        switch (*in) {
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            t = t * 10 + (*in - '0');
            break;
        case 'Y':case 'y':
            t *= 53;
        case 'W':case 'w':
            t *= 7;
        case 'D':case 'd':
            t *= 24;
        case 'H':case 'h':
            t *= 60;
        case 'M':case 'm':
            t *= 60;
        case 'S':case 's':
            tt += t;
            t = 0;
            break;
        default:
            ifBLIN_QX0("Illegal time %c(%02X)", *in, *in & 0xFF);
            tt = 0;
            goto out;
    }   }
    tt += t;
out:
    return(tt);
#   undef blin_internal_flags
}

int
get_t_mask(deleted_pid *parm, babolo_opts *bos) {
    const char *in;
    int res = 0;

#   define blin_internal_flags (parm->flags & BLIN_MASK)
    if  (!(in = babolo_getoptsarg(bos))) {
        ifBLIN_QW0("getoptsarg -%c", bos->c);
        goto out;
    }
    for (; in && *in; ++in) {
        switch (*in) {
        case 'A':case 'a': res |= USE_A; break;
        case 'M':case 'm': res |= USE_M; break;
        case 'C':case 'c': res |= USE_C; break;
        case 'B':case 'b': res |= USE_B; break;
        case 'L':case 'l': res |= USE_L; break;
        case 'P':case 'p': res |= USE_P; break;
        case 'O':case 'o': res |= USE_O; break;
        case 'G':case 'g': res |= USE_G; break;
        case 'R':case 'r': res |= USE_R; break;
        case 'N':case 'n': res |= NO_RM; break;
        default:
            ifBLIN_QX0("Illegal mask %c(%02X)", *in, *in & 0xFF);
            res = 0;
            goto out;
    }   }
out:
    return(res);
#   undef blin_internal_flags
}

static void
usage(int ex) {
    fprintf( stderr
           , "deleted @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "Usage:\n"
             "deleted -t TimeToLive[ -n Nice][ -f mask][ -d mask][ -l mask][ -p pidfile] directory\n"
             "    where mask is some word from [amclp]*\n"
             "    for files (-f), directories (-d) and links(-l)\n"
             "        a - use atime\n"
             "        m - use mtime\n"
             "        c - use ctime\n"
             "        b - use birthtime\n"
             "        l - use lock\n"
             "        p - use pointed by symlink\n"
             "    You can use instead -t:\n"
             "        -F TimeToLiveFiles\n"
             "        -D TimeToLiveDirs\n"
             "        -L TimeToLiveLinks\n"
             " or\n"
             "deleted -q directory\n"
             "    to query running deleted daemon\n"
           );
    exit(ex);
}

static void
huphandler(int signal) {return;}

int
main(int argc, char **argv) {
    struct sigaction act = {{huphandler}, 0, {{0}}};
    const char *o, *flar = "d:D:f:F:l:L:n:p:qt:v?";
    int c, prio, pidfd = -1, ex = EX_OK;
    deleted_pid parm;
    babolo_opts *bos;

    bzero(&parm, sizeof(deleted_pid));
    parm.flags = blin_internal_flags & BLIN_MASK;
    parm.dmask.mask = USE_M | USE_C | USE_B | USE_L;
    parm.fmask.mask = USE_M | USE_C | USE_B | USE_L | USE_A;
    parm.lmask.mask = USE_M | USE_C | USE_B | USE_L | USE_A | USE_P;
    parm.fmask.mod = parm.dmask.mod = parm.lmask.mod = 0444;
    parm.pidname = NAMEPID;
    if  (!(bos = babolo_openopts(parm.flags | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        EXOUT(EX_SOFTWARE);
    }
    if  (!!babolo_setopts(bos, parm.flags | Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_USAGE;
    while (!!(c = babolo_getopts(bos))) {
        switch (c) {
        case 'd': if  (0 >= (parm.dmask.mask = get_t_mask(&parm, bos))) {
                      ifBLIN_QW0("get_t_mask d -%c %08X", bos->c, parm.dmask.mask);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'D': if  (0 >= (parm.dmask.ttl = get_time(&parm, bos))) {
                      ifBLIN_QW0("get_time d -%c %d", bos->c, parm.dmask.ttl);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'f': if  (0 >= (parm.fmask.mask = get_t_mask(&parm, bos))) {
                      ifBLIN_QW0("get_t_mask f -%c %08X", bos->c, parm.fmask.mask);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'F': if  (0 >= (parm.fmask.ttl = get_time(&parm, bos))) {
                      ifBLIN_QW0("get_time f -%c %d", bos->c, parm.fmask.ttl);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'l': if  (0 >= (parm.lmask.mask = get_t_mask(&parm, bos))) {
                      ifBLIN_QW0("get_t_mask f -%c %08X", bos->c, parm.lmask.mask);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'L': if  (0 >= (parm.lmask.ttl = get_time(&parm, bos))) {
                      ifBLIN_QW0("get_time f -%c %d", bos->c, parm.lmask.ttl);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'n': parm.prio = babolo_getoptunum(bos, 0);
                ; if  (!!errno) {
                      ifBLIN_QW0("getoptunum -%c", bos->c);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'p': if  (!(parm.pidname = babolo_getoptsarg(bos))) {
                      ifBLIN_QW0("getoptsarg -%c", bos->c);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'q': parm.flags |= DELETED_QUERY;
                ; break;
        case 't': parm.lmask.ttl = parm.dmask.ttl = parm.fmask.ttl = get_time(&parm, bos);
                ; if  (0 >= parm.lmask.ttl) {
                      ifBLIN_QW0("get_time f -%c %d", bos->c, parm.lmask.ttl);
                      EXOUT(EX_USAGE);
                  }
                ; break;
        case 'v': BLIN_VERBOSE(parm.flags);
                ; break;
        case '=': ex = EX_OK;
                ; switch(babolo_getoptsarg(bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP); goto out;
                  case '1': babolo_optest(flar);     goto out;
                  case '2': babolo_optext(bos);      goto out;
                  case '3': babolo_dumpopts(bos);    goto out;
                  }
                ; break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    if  (!(o = babolo_getargs(bos))) {
        ifBLIN_QW0("getargs");
        EXOUT(EX_USAGE);
    }
    parm.id = MY_ID;
    parm.version = VMAJOR;
    parm.dpss = sizeof(parm);
    /******************************************/
    /* o may be absolute path. or relative... */
    /****   it is what seteuid for         ****/
    /******************************************/
    /****/  if  (!!chroot(o)) {           /****/
    /****/      ifBLIN_QW0("chroot");     /****/
    /****/      EXOUT(EX_UNAVAILABLE);    /****/
    /****/  }                             /****/
    /****/  if  (!!setuid(getuid())) {    /****/
    /****/      ifBLIN_QW0("setuid");     /****/
    /****/      EXOUT(EX_UNAVAILABLE);    /****/
    /****/  }                             /****/
    /******************************************/
    /****   END seteuid bit                ****/
    /******************************************/
    if  (!!chdir("/")) {
        ifBLIN_QW0("chdir /");
        EXOUT(EX_UNAVAILABLE);
    }
    parm.pid = getpid();
    snprintf(parm.cpid, C_PID_LEN, "%d\n", parm.pid);
    if  (parm.flags & DELETED_QUERY) {
        ex = query(&parm);
        goto out;
    } else if (!parm.fmask.ttl && !parm.dmask.ttl && !parm.lmask.ttl) {
        ifBLIN_QX0("some TimeToLive must be specified");
        usage(EX_USAGE);
    }
    prio = prio_set(&parm);
    if  (parm.prio < prio) {
        ifBLIN_QX0("Nice %d impossible, %d used instead", parm.prio, prio);
        parm.prio = prio;
    }
    if  (0 > (pidfd = open(parm.pidname, O_RDWR | O_CREAT, 0600))) {
        ifBLIN_QW0("open/create %s", parm.pidname);
        EXOUT(EX_IOERR);
    }
    parm.conf = NULL;
    /* If two or more deleted starts almoust simultaniesly for this dir, one loop can be unsuccessful*/
    while (0 <= pidfd) {
    ;   if  (!flock(pidfd, LOCK_EX | LOCK_NB)) {                      /* If no other deleted running */
            struct stat sb, sbf;

    ;       ifBLIN_QX1("First deleted started");
    ;       if  (!!lstat(parm.pidname, &sb)) {
    ;           ifBLIN_QW0("lstat");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       if  (!!fstat(pidfd, &sbf)) {
    ;           ifBLIN_QW0("fstat");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       if  ((!S_ISREG(sb.st_mode)) || (sb.st_dev != sbf.st_dev) || (sb.st_ino != sbf.st_ino)) {
    ;           ifBLIN_QX0("!!! Substitute attack !!! -> %s", parm.pidname);
    ;           EXOUT(EX_TEMPFAIL);
    ;       }
    ;       if  (!!ftruncate(pidfd, sizeof(deleted_pid))) {
    ;           ifBLIN_QW0("ftruncate");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       parm.conf = (deleted_pid *)mmap( NULL
                                           , sizeof(deleted_pid)
                                           , PROT_READ | PROT_WRITE
                                           , MAP_SHARED
                                           , pidfd
                                           , (off_t)0
                                           );
    ;       if  (MAP_FAILED == (void*)parm.conf) {
    ;           ifBLIN_QW0("mmap");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       bcopy(&parm, parm.conf, sizeof(deleted_pid));
    ;       parm.conf->pidname = NULL;
    ;       parm.conf->conf = NULL;
    ;       if  (!!flock(pidfd, LOCK_SH)) {
    ;           ifBLIN_QW0("flock shared after exclusive");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       if  (!!sigaction(SIGHUP, &act, NULL)) {
    ;           ifBLIN_QW0("sigaction");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       main_loop(&parm, prio);
    ;       break;
    ;   } else if (!flock(pidfd, LOCK_SH | LOCK_NB)) {                /* If no other deleted running */
    ;       ifBLIN_QW1("Not first deleted started");
    ;       parm.conf = (deleted_pid *)mmap( NULL
                                           , sizeof(deleted_pid)
                                           , PROT_READ | PROT_WRITE
                                           , MAP_SHARED
                                           , pidfd
                                           , (off_t)0
                                           );
    ;       if  (MAP_FAILED == (void*)parm.conf) {
    ;           ifBLIN_QW0("mmap");
    ;           EXOUT(EX_UNAVAILABLE);
    ;       }
    ;       if  (parm.conf->id != parm.id) {
    ;           ifBLIN_QX0( "%s belongs to program %016X <> %016X"
                          , parm.pidname, parm.conf->id, parm.id
                          );
    ;           EXOUT(EX_DATAERR);
    ;       }
    ;       if  (parm.conf->version != parm.version) {
    ;           ifBLIN_QX0( "%s belongs to version %u <> %u"
                          , parm.pidname, parm.conf->version, parm.version
                          );
    ;           EXOUT(EX_DATAERR);
    ;       }
    ;       if  (parm.conf->dpss != parm.dpss) {
    ;           ifBLIN_QX0("%s belongs to arch %u <> %u", parm.pidname, parm.conf->dpss, parm.dpss);
    ;           EXOUT(EX_DATAERR);
    ;       }
    ;       bcopy(&parm.fmask, &parm.conf->fmask, sizeof(aim));
    ;       bcopy(&parm.dmask, &parm.conf->dmask, sizeof(aim));
    ;       bcopy(&parm.lmask, &parm.conf->lmask, sizeof(aim));
    ;       parm.conf->flags = parm.flags;
    ;       if  (parm.conf->prio > parm.prio) {
    ;           ifBLIN_QX1("Second deleted substitutes");
    ;           parm.conf->prio = parm.prio;
    ;           if  (!!kill(parm.conf->pid, SIGHUP)) {
    ;               ifBLIN_QX0("kill");
    ;               if  ((errno != ESRCH) && (errno != EPERM)) {
    ;                   EXOUT(EX_UNAVAILABLE);
    ;           }   }
    ;       } else {
    ;           ifBLIN_QX1("Second deleted corrects");
    ;           parm.conf->prio = parm.prio;
    ;           if  (!kill(parm.conf->pid, SIGHUP)) {
    ;               break;
    ;           } else {
    ;               ifBLIN_QX0("kill");
    ;               if  ((errno != ESRCH) && (errno != EPERM)) {
    ;                   EXOUT(EX_UNAVAILABLE);
    ;       }   }   }
    ;   } else if (EWOULDBLOCK != errno) {
    ;       ifBLIN_QW0("flock %s", parm.pidname);
    ;       EXOUT(EX_UNAVAILABLE);
    ;   }
    ;   if  (parm.conf) munmap(parm.conf, sizeof(deleted_pid));
    ;   parm.conf = NULL;
    ;   if  (0 <= pidfd) flock(pidfd, LOCK_UN | LOCK_NB);
    ;   ifBLIN_QX1("Next try");
    ;   usleep(30000);
    }
out:
    if  (parm.conf) munmap(parm.conf, sizeof(deleted_pid));
    parm.conf = NULL;
    if  (0 <= pidfd) flock(pidfd, LOCK_UN | LOCK_NB);
    exit(ex);
}
